#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import threading
import unittest

import pytensorpipe as tp


class TestTensorpipe(unittest.TestCase):
    def test_read_write(self):
        context = tp.Context()
        context.register_transport(0, "tcp", tp.create_uv_transport())
        create_shm_transport = getattr(tp, "create_shm_transport", None)
        if create_shm_transport is not None:
            context.register_transport(-1, "shm", create_shm_transport())
        context.register_channel(0, "basic", tp.create_basic_channel())
        create_cma_channel = getattr(tp, "create_cma_channel", None)
        if create_cma_channel is not None:
            context.register_channel(-1, "cma", create_cma_channel())

        # We must keep a reference to it, or it will be destroyed early.
        server_pipe = None

        listener: tp.Listener = context.listen(["tcp://127.0.0.1"])

        write_completed = threading.Event()

        def on_connection(pipe: tp.Pipe) -> None:
            global server_pipe
            payload = tp.OutgoingPayload(b"Hello ", b"a greeting")
            tensor = tp.OutgoingTensor(b"World!", b"a place")
            message = tp.OutgoingMessage(b"metadata", [payload], [tensor])
            pipe.write(message, on_write)
            server_pipe = pipe

        def on_write() -> None:
            write_completed.set()

        listener.listen(on_connection)

        client_pipe: tp.Pipe = context.connect(listener.get_url("tcp"))

        received_payloads = None
        received_tensors = None
        read_completed = threading.Event()

        def on_read_descriptor(message: tp.IncomingMessage) -> None:
            nonlocal received_payloads, received_tensors
            self.assertEqual(message.metadata, bytearray(b"metadata"))
            received_payloads = []
            for payload in message.payloads:
                self.assertEqual(payload.metadata, bytearray(b"a greeting"))
                received_payloads.append(bytearray(payload.length))
                payload.buffer = received_payloads[-1]
            received_tensors = []
            for tensor in message.tensors:
                self.assertEqual(tensor.metadata, bytearray(b"a place"))
                received_tensors.append(bytearray(tensor.length))
                tensor.buffer = received_tensors[-1]
            client_pipe.read(message, on_read)

        def on_read() -> None:
            read_completed.set()

        client_pipe.read_descriptor(on_read_descriptor)

        write_completed.wait()
        read_completed.wait()

        self.assertEqual(received_payloads, [bytearray(b"Hello ")])
        self.assertEqual(received_tensors, [bytearray(b"World!")])

        # Due to a current limitation we're not releasing the GIL when calling
        # the context's destructor, which implicitly calls join, which may fire
        # some callbacks that also try to acquire the GIL and thus deadlock.
        # So, for now, we must explicitly call join.
        # See https://github.com/pybind/pybind11/issues/1446.
        context.join()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import json
import sys

if __name__ == "__main__":
    if len(sys.argv) < 1:
        raise RuntimeError()
    if len(sys.argv) != 4:
        print(
            f"Usage: {sys.argv[0]} [first report] [second report] [supposed to work]",
            file=sys.stderr,
        )
        sys.exit(0)

    with open(sys.argv[1], "rb") as f:
        first_report = json.load(f)
    with open(sys.argv[2], "rb") as f:
        second_report = json.load(f)
    supposed_to_work = int(sys.argv[3])

    worked_in_practice = (
        first_report["syscall_success"] == 1 and second_report["syscall_success"] == 1
    )
    if worked_in_practice != supposed_to_work:
        raise RuntimeError(
            f"The syscall didn't behave as the test expected it to. It "
            f"{'succeeded' if worked_in_practice else 'failed'} whereas it was "
            f"supposed to {'succeed' if supposed_to_work else 'fail'}."
        )

    detected_as_working = (
        first_report["viability"] == 1
        and second_report["viability"] == 1
        and first_report["device_descriptor"] == second_report["device_descriptor"]
    )
    if detected_as_working != worked_in_practice:
        print(
            f"The CMA autodetection didn't correctly predict the behavior of the "
            f"syscall. It determined it would "
            f"{'succeed' if detected_as_working else 'fail'} whereas it actually "
            f"{'succeeded' if worked_in_practice else 'failed'}.",
            file=sys.stderr,
        )
        sys.exit(1)

    sys.exit(0)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import os
import subprocess
import sys
from pathlib import Path

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext


class CMakeBuild(build_ext):
    def run(self):
        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        source_path = Path(__file__).parent.resolve()
        output_path = Path(self.get_ext_fullpath(ext.name)).parent.resolve()
        build_type = "Debug" if self.debug else "Release"

        cmake_cmd = [
            "cmake",
            f"{source_path}",
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={output_path}",
            f"-DPYTHON_EXECUTABLE={sys.executable}",
            f"-DCMAKE_BUILD_TYPE={build_type}",
            "-DCMAKE_C_COMPILER=clang-6.0",
            "-DCMAKE_CXX_COMPILER=clang++-6.0",
            "-DCMAKE_POSITION_INDEPENDENT_CODE=true",
            "-DTP_BUILD_PYTHON=true",
        ]

        for opt in os.environ:
            if opt.startswith("TP_"):
                cmake_cmd.append(f"-D{opt}={os.environ[opt]}")

        make_cmd = ["make", "-j", "pytensorpipe"]

        subprocess.check_call(cmake_cmd, cwd=self.build_temp)
        subprocess.check_call(make_cmd, cwd=self.build_temp)


setup(
    name="tensorpipe",
    version="0.0.0",
    author="Facebook AI Research",
    ext_modules=[Extension("pytensorpipe", sources=[])],
    cmdclass={"build_ext": CMakeBuild},
    zip_safe=False,
)
